#include <string.h>

#include <mutex>
#include <condition_variable>
#include <thread>

#include <unistd.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/epoll.h>
#include <netinet/tcp.h>

#include <a8/a8.h>
#include <a8/asynctcpclient.h>
#include <a8/ioloop.h>

static const int DEFAULT_MAX_RECV_BUFFERSIZE = 1024 * 64;

namespace a8
{

    AsyncTcpClient::AsyncTcpClient()
    {
        ref_count = 1;
        send_buffer_mutex_ = new std::mutex();
    }

    AsyncTcpClient::~AsyncTcpClient()
    {
        if (!list_empty(&connect_timer_attacher.timer_list_)) {
            abort();
        }
        Close();
        A8_SAFE_DELETE(send_buffer_mutex_);
    }

    void AsyncTcpClient::Open()
    {
        if (!IsActive()) {
            SetActive(true);
        }
    }

    void AsyncTcpClient::Close()
    {
        if (IsActive()) {
            SetActive(false);
        }
    }

    bool AsyncTcpClient::IsActive()
    {
        return socket_ != a8::INVALID_SOCKET;
    }

    bool AsyncTcpClient::Connected()
    {
        return connected_;
    }

    void AsyncTcpClient::SendBuff(const char* buff, unsigned int bufflen)
    {
        if (bufflen > 0) {
            a8::SendQueueNode* p = (a8::SendQueueNode*)malloc(sizeof(a8::SendQueueNode));
            memset(p, 0, sizeof(SendQueueNode));
            p->buff = (char*)malloc(bufflen);
            memmove(p->buff, buff, bufflen);
            p->bufflen = bufflen;
            send_buffer_mutex_->lock();
            if (bot_node_) {
                bot_node_->next = p;
                bot_node_ = p;
            }else{
                top_node_ = p;
                bot_node_ = p;
            }
            if (!sending_) {
                NotifyEpollSend();
            }
            send_buffer_mutex_->unlock();
        }
    }

    void AsyncTcpClient::DoConnect()
    {
        connected_ = true;
        if (on_connect) {
            on_connect(this);
        }
    }

    void AsyncTcpClient::SetEpollFd(int epoll_fd)
    {
        epoll_fd_ = epoll_fd;
    }

    void AsyncTcpClient::DoRecv()
    {
        if (socket_ == -1) {
            return;
        }
        char recvbuf[DEFAULT_MAX_RECV_BUFFERSIZE];
        while (true) {
            int ret = ::recv(socket_, recvbuf, DEFAULT_MAX_RECV_BUFFERSIZE, 0);
            if (ret < 0) {
                if (errno != EAGAIN) {
                    Close();
                    return;
                } else {
                }
                break;
            } else if (ret == 0) {
                Close();
                return;
            } else {
                if (on_socketread) {
                    on_socketread(this, recvbuf, ret);
                }
                if (ret < DEFAULT_MAX_RECV_BUFFERSIZE) {
                    break;
                }
            }
        }
    }

    void AsyncTcpClient::DoSend()
    {
        if (socket_ == -1) {
            return;
        }
        if (!work_node_) {
            send_buffer_mutex_->lock();
            work_node_ = top_node_;
            top_node_ = nullptr;
            bot_node_ = nullptr;
            send_buffer_mutex_->unlock();
        }
        if (!connected_) {
            int error = 0;
            socklen_t len = sizeof(error);
            ::getsockopt(socket_, SOL_SOCKET, SO_ERROR, &error, (socklen_t *)&len);
            if( !error ) {
                if (!work_node_) {
                    struct epoll_event ev;
                    ev.data.fd = socket_;
                    ev.events = EPOLLIN | EPOLLRDHUP;
                    ev.data.ptr = this;
                    ::epoll_ctl(epoll_fd_, EPOLL_CTL_MOD, socket_, &ev);
                }
                DoConnect();
            } else {
                connected_ = false;
                if (socket_ != a8::INVALID_SOCKET) {
                    ::shutdown(socket_, 2);
                    ::close(socket_);
                }
                socket_ = a8::INVALID_SOCKET;
                ClearSendBuff();
                if (on_error) {
                    on_error(this, error);
                }
                return;
            }
        }
        if (work_node_) {
            AsyncSend();
        }
    }

    void AsyncTcpClient::DoError()
    {
        connected_ = false;
        if (on_disconnect) {
            on_disconnect(this);
        }
    }

    void AsyncTcpClient::SetActive(bool active)
    {
        if (active) {
            a8::IoLoop::Instance()->PostAsyncConnect(this, timeout_ms);
        } else {
            a8::IoLoop::Instance()->PostAsyncClose(this);
        }
    }

    void AsyncTcpClient::DoAsyncConnect()
    {
        socket_ = ::socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
        if (a8::INVALID_SOCKET == socket_) {
            if (on_error) {
                on_error(this, errno);
            }
            return;
        }
        //set nodelay
        {
          int flag = 1;
          int ret = ::setsockopt(socket_,
                                 IPPROTO_TCP,
                                 TCP_NODELAY,
                                 (char *)&flag,
                                 sizeof(flag));
          assert(ret >= 0);
          if (ret < 0) {
            abort();
          }
        }
        //set nonblock
        {
          int flags = 0;
          flags = ::fcntl(socket_, F_GETFL, 0);
          ::fcntl(socket_, F_SETFL, flags|O_NONBLOCK);
        }
        sockaddr_in sa;
        memset(&sa, 0, sizeof(sa));
        sa.sin_family = AF_INET;
        sa.sin_addr.s_addr = inet_addr(remote_address.c_str());
        sa.sin_port = htons(remote_port);
        int ret = ::connect(socket_, (sockaddr*)&sa, sizeof(sa));
        if (ret == 0) {
            //add epoll
            {
                struct epoll_event ev;
                ev.data.fd = socket_;
                ev.events = EPOLLIN | EPOLLRDHUP | EPOLLERR;
                ev.data.ptr = this;
                int n = ::epoll_ctl(epoll_fd_, EPOLL_CTL_ADD, socket_, &ev);
                assert(n == 0);
                if (n != 0) {
                    abort();
                }
            }
            DoConnect();
        } else if (ret < 0) {
            if (errno == EINPROGRESS) {
                //add epoll
                {
                    struct epoll_event ev;
                    ev.data.fd = socket_;
                    ev.events = EPOLLIN | EPOLLOUT | EPOLLRDHUP | EPOLLERR;
                    ev.data.ptr = this;
                    int n = ::epoll_ctl(epoll_fd_, EPOLL_CTL_ADD, socket_, &ev);
                    assert(n == 0);
                    if (n != 0) {
                        abort();
                    }
                }
            } else {
                connected_ = false;
                if (on_error) {
                    on_error(this, errno);
                }
                ::close(socket_);
                socket_ = a8::INVALID_SOCKET;
                return;
            }
        }
    }

    void AsyncTcpClient::DoAsyncClose()
    {
        sending_ = false;
        connected_ = false;
        if (socket_ != a8::INVALID_SOCKET) {
            ::shutdown(socket_, 2);
            ::close(socket_);
        }
        socket_ = a8::INVALID_SOCKET;
        ClearSendBuff();
    }

    void AsyncTcpClient::ClearSendBuff()
    {
        send_buffer_mutex_->lock();
        a8::ClearSendQueue(top_node_);
        top_node_ = nullptr;
        bot_node_ = nullptr;
        send_buffer_mutex_->unlock();
        a8::ClearSendQueue(work_node_);
        work_node_ = nullptr;
    }

    void AsyncTcpClient::NotifyEpollSend()
    {
        sending_ = true;
        struct epoll_event ev;
        ev.data.fd = socket_;
        ev.events = EPOLLIN | EPOLLOUT | EPOLLRDHUP;
        ev.data.ptr = this;
        ::epoll_ctl(epoll_fd_, EPOLL_CTL_MOD, socket_, &ev);
    }

    void AsyncTcpClient::AsyncSend()
    {
        while (work_node_) {
            int sentbytes = ::send(socket_,
                                   work_node_->buff + work_node_->sent_bytes,
                                   work_node_->bufflen - work_node_->sent_bytes,
                                   0);
            if (sentbytes <= 0) {
                auto err_code = errno;
                if (err_code == EAGAIN || err_code == EWOULDBLOCK) {
                    break;
                } else {
                    Close();
                    break;
                }
            }
            work_node_->sent_bytes += sentbytes;
            if (work_node_->sent_bytes >= work_node_->bufflen) {
                a8::SendQueueNode *pdelnode = work_node_;
                work_node_ = work_node_->next;
                if (!work_node_) {
                    send_buffer_mutex_->lock();
                    if (top_node_) {
                        work_node_ = top_node_;
                        top_node_ = nullptr;
                        bot_node_ = nullptr;
                    }
                    if (!work_node_) {
                        sending_ = false;
                        struct epoll_event ev;
                        ev.data.fd = socket_;
                        ev.events = EPOLLIN | EPOLLRDHUP;
                        ev.data.ptr = this;
                        ::epoll_ctl(epoll_fd_, EPOLL_CTL_MOD, socket_, &ev);
                    }
                    send_buffer_mutex_->unlock();
                }
                if (pdelnode->buff) {
                    free(pdelnode->buff);
                }
                free(pdelnode);
            }
        }
    }

}
